import numpy as np
import argparse
import matplotlib.pyplot as plt

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--value', default='variables', type=str, help='Choices variables, samples')
    parser.add_argument('--noise', default='gauss', type=str, help='Choices none, gauss, gumbel, uniform')
    parser.add_argument('--noise_std', default=0.01, type=float, help='Noise magnitude')
    parser.add_argument('--noise_effect', default='signal', type=str, help='Where the noise is applied ')
    parser.add_argument('--trans_clos', default='FW', type=str, help='Choices FW and zero')
    parser.add_argument('--sparsity', default=0.3, type=float, help='Probability of data being nonzero at vertex v')

    parser.add_argument('--weight_bounds', default=[0.4, 0.8], nargs='+', type=float, help='initialization of weighted adjacency matrix')
    parser.add_argument('--edges', default=2, type=int, help='graph has k * d edges')

    parser.add_argument('--samples', default=[400], nargs='+', type=int, help='number of samples')
    parser.add_argument('--nodes', default=[10, 20, 30, 40, 50], nargs='+', type=int, help='number of graph vertices to consider')# [5, 10, 15, 20, 25]
    parser.add_argument('--fixSup', default='False', type=str, help='Whether to fix the support of the spectrum')
    parser.add_argument('--methods', default=['moebius', 'notears', 'nocurl', 'GES', 'FGS', 'MMPC', 'CAM'], nargs='+', type=str, help='methods to compare')
    parser.add_argument('--transformation', default='None', type=str, help='Whether to normalize/standardize the given signals')
    parser.add_argument('--runs', default=10, type=int)
    args = parser.parse_args()

    samples = [400] #[200, 400, 600, 800, 1000] if(args.value == 'samples') else [400]
    variables =  args.nodes # if(args.value == 'variables') else [10] 
    noise = args.noise
    runs = args.runs 
    (a, b) = tuple(args.weight_bounds)
    k = args.edges
    x_axis = samples if len(samples) > 1 else variables
    methods = args.methods

    # finding the name of the output files according to the experimental settings 
    dic = vars(args)
    filename = ''
    for key in dic.keys():
        if(key not in ['methods', 'variables'] and dic[key]!= parser.get_default(key)):
            filename += '{}_{}_'.format(key, dic[key])
    filename = filename if len(filename) > 0 else 'default'

    color_methods = {
        'moebius': 'C4', 
        'notears': 'C2', 
        'nocurl' : 'C5', 
        'GES': 'C6', 
        'MMPC': 'C7', 
        'CAM' : 'C8', 
        'FGS' : 'C3'
    }
    label_methods = {
        'moebius': 'Möbius', 
        'notears': 'NOTEARS', 
        'nocurl' : 'DAG-NoCurl', 
        'GES': 'GES', 
        'MMPC': 'MMPC', 
        'CAM' : 'CAM', 
        'FGS' : 'fGES'
    }

    avg = {}
    std = {}
    for key in methods:
        avg[key] = []
        std[key] = []

    with open('results/{}.csv'.format(filename), 'r') as f:
        for line in f:
            info = line.split(',')
            for method in methods:
                if(info[0] == 'Acc {} is'.format(method)):
                    avg[method].append([float(info[1]), float(info[2]), float(info[3]), float(info[4])])
                elif(info[0] == 'Std {} is'.format(method)):
                    std[method].append([float(info[1]), float(info[2]), float(info[3]), float(info[4])])
        
    for method in methods:
        avg[method] = np.array(avg[method])
        std[method] = np.array(std[method])

    from matplotlib import font_manager
    font_dirs = ['fonts/']
    font_files = font_manager.findSystemFonts(fontpaths=font_dirs)
    for font_file in font_files:
        font_manager.fontManager.addfont(font_file)

    with plt.style.context('ggplot'):
        plt.rcParams['font.family'] = 'gillsans'
        plt.rcParams['xtick.color'] = 'black'
        plt.rcParams['ytick.color'] = 'black'

        for i, label in enumerate(['SHD', 'TPR', 'NNZ', 'NMSE']):
            plt.figure()
            for method in methods:
                if(len(avg[method] > 0)):
                    plt.plot(x_axis, avg[method][:, i], label = label_methods[method], color=color_methods[method])
                    plt.fill_between(x_axis, avg[method][:, i] - std[method][:, i], avg[method][:, i] + std[method][:, i], color=color_methods[method], alpha=.1)

            # if(label == 'TPR'):
            #     plt.ylim(0, 1)
            plt.ylabel(label, fontsize=20)
            plt.xlabel('Number of sample sizes' if args.value=='samples' else 'Number of vertices', fontsize=20)
            plt.xticks(x_axis, fontsize=18)
            plt.yticks(fontsize=18)
            plt.grid(axis='x', linewidth=0.1)
            plt.legend(frameon=False, fontsize=18)
            plt.tight_layout()
            plt.savefig('plots/plot_{}_{}.png'.format(filename, label))

