import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--value', default='variables', type=str, help='Choices variables, samples')
    parser.add_argument('--noise', default='gauss', type=str, help='Choices none, gauss, gumbel, uniform')
    parser.add_argument('--noise_std', default=0.01, type=float, help='Noise magnitude')
    parser.add_argument('--noise_effect', default='signal', type=str, help='Where the noise is applied ')
    parser.add_argument('--trans_clos', default='FW', type=str, help='Choices FW and zero')
    parser.add_argument('--sparsity', default=0.3, type=float, help='Probability of data being nonzero at vertex v')

    parser.add_argument('--weight_bounds', default=[0.4, 0.8], nargs='+', type=float, help='initialization of weighted adjacency matrix')
    parser.add_argument('--edges', default=2, type=int, help='graph has k * d edges')

    parser.add_argument('--samples', default=[400], nargs='+', type=int, help='number of samples')
    parser.add_argument('--nodes', default=[10, 20, 30, 40, 50], nargs='+', type=int, help='number of graph vertices to consider')# [5, 10, 15, 20, 25]
    parser.add_argument('--fixSup', default='False', type=str, help='Whether to fix the support of the spectrum')
    parser.add_argument('--methods', default=['moebius', 'notears', 'nocurl', 'GES', 'FGS', 'MMPC', 'CAM'], nargs='+', type=str, help='methods to compare')
    parser.add_argument('--transformation', default='None', type=str, help='Whether to normalize/standardize the given signals')
    parser.add_argument('--runs', default=10, type=int)
    args = parser.parse_args()


    methods = args.methods

    # finding the name of the output files according to the experimental settings 
    dic = vars(args)
    filename = ''
    for key in dic.keys():
        if(key not in ['methods', 'variables'] and dic[key]!= parser.get_default(key)):
            filename += '{}_{}_'.format(key, dic[key])
    filename = filename if len(filename) > 0 else 'default'

    avg = {}
    std = {}
    for key in methods:
        avg[key] = []
        std[key] = []

    best_tpr = 0
    with open('results/{}.csv'.format(filename), 'r') as f:
        for line in f:
            info = line.split(',')
            for method in methods:
                if(info[0] == 'Acc {} is'.format(method)):
                    avg[method].append([float(info[1]), float(info[2]), float(info[3]), float(info[4])])
                elif(info[0] == 'Std {} is'.format(method)):
                    std[method].append([float(info[1]), float(info[2]), float(info[3]), float(info[4])])
                if len(avg[method]) == 4:
                    best_tpr = max(avg[method][3][1], best_tpr)
            
    result = ''
    for method in methods:
        if(avg[method][3][1] < best_tpr):
            result += '&  $    {:.2f}\pm{:.2f} $  '.format(avg[method][3][1], std[method][-2][1])
        else:
            result += '&  $\\bm' + '{' + '{:.2f}\pm{:.2f}'.format(avg[method][3][1], std[method][-2][1]) + '}$  '
        
    result += '\\\\ \n'
    with open('results/table.tex', 'a') as f:
        f.write(result)