import networkx as nx
import notears.notears.utils as notears_utils
from notears.notears.linear import * 
from pycausal.pycausal import pycausal as pc # pycausal package is used in fast greedy equivalence search algorithm (fGES or FGS)
from NoCurl.BPR import *
from NoCurl.main_efficient import *
import cdt 
cdt.SETTINGS.rpath = 'C:/Program Files/R/R-4.2.1/bin/Rscript'

methods = ['moebius', 'notears', 'nocurl', 'FGS', 'GES', 'MMPC', 'CAM']

def run_real():    
    data, graph = cdt.data.load_dataset('sachs')

    B_true = nx.to_numpy_array(graph)

    X = data.to_numpy()

    print(X.shape)
    
    with open('results/sachs.csv', 'a') as f:
        for method in methods:
            if method == 'moebius':
                W_est = notears_linear(X, lambda1=0.005, loss_type='moebius', trans_clos='FW') # for spectral sem λ=1, for linear λ=0.001
                B_est = W_est != 0
            elif method == 'notears':
                W_est = notears_linear(X, lambda1=0, loss_type='l2')
                B_est = W_est != 0
            else:
                tmp_args = get_args()
                bpr = BPR.BPR(tmp_args, pc)
                W_est, h, alpha, rho = bpr.fit(X, method)
                B_est = W_est != 0 if method == 'nocurl' else W_est

            # nmse = np.linalg.norm(W_est - W_true) / np.linalg.norm(W_true)
            acc = notears_utils.count_accuracy(B_true, B_est)
            print("Acc {} is, SHD {:.3f}, ACC {:.3f}, NNZ {:.3f}".format(method, acc['shd'], acc['tpr'], acc['nnz']))
            f.write("Acc {} is, SHD {:.3f}, ACC {:.3f}, NNZ {:.3f}\n".format(method, acc['shd'], acc['tpr'], acc['nnz']))

if __name__ == '__main__':
    pc = pc()
    pc.start_vm()
    run_real()
    pc.stop_vm()